/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package com.datastax.driver.core.policies;

import com.datastax.driver.core.Cluster;
import com.datastax.driver.core.ConsistencyLevel;
import com.datastax.driver.core.Statement;
import com.datastax.driver.core.WriteType;
import com.datastax.driver.core.exceptions.DriverException;

/**
 * A retry policy that never retries (nor ignores).
 *
 * <p>All of the methods of this retry policy unconditionally return {@link
 * RetryPolicy.RetryDecision#rethrow()}. If this policy is used, retry logic will have to be
 * implemented in business code.
 */
public class FallthroughRetryPolicy implements RetryPolicy {

  public static final FallthroughRetryPolicy INSTANCE = new FallthroughRetryPolicy();

  private FallthroughRetryPolicy() {}

  /**
   * {@inheritDoc}
   *
   * <p>This implementation always returns {@code RetryDecision.rethrow()}.
   */
  @Override
  public RetryDecision onReadTimeout(
      Statement statement,
      ConsistencyLevel cl,
      int requiredResponses,
      int receivedResponses,
      boolean dataRetrieved,
      int nbRetry) {
    return RetryDecision.rethrow();
  }

  /**
   * {@inheritDoc}
   *
   * <p>This implementation always returns {@code RetryDecision.rethrow()}.
   */
  @Override
  public RetryDecision onWriteTimeout(
      Statement statement,
      ConsistencyLevel cl,
      WriteType writeType,
      int requiredAcks,
      int receivedAcks,
      int nbRetry) {
    return RetryDecision.rethrow();
  }

  /**
   * {@inheritDoc}
   *
   * <p>This implementation always returns {@code RetryDecision.rethrow()}.
   */
  @Override
  public RetryDecision onUnavailable(
      Statement statement,
      ConsistencyLevel cl,
      int requiredReplica,
      int aliveReplica,
      int nbRetry) {
    return RetryDecision.rethrow();
  }

  /**
   * {@inheritDoc}
   *
   * <p>This implementation always returns {@code RetryDecision.rethrow()}.
   */
  @Override
  public RetryDecision onRequestError(
      Statement statement, ConsistencyLevel cl, DriverException e, int nbRetry) {
    return RetryDecision.rethrow();
  }

  @Override
  public void init(Cluster cluster) {
    // nothing to do
  }

  @Override
  public void close() {
    // nothing to do
  }
}
